use defmt::*;
use {defmt_rtt as _, panic_probe as _};
use heapless::String;

use embassy_futures::select::*;
use embassy_time::{Duration, Timer};
use embassy_rp::gpio::{self, Pull, Flex};
use embassy_rp::i2c;

fn append_hex1<const N : usize>(s : &mut String<N>, num : u8) -> Result<(), ()> {
    let ch = if num < 10 {
        ('0' as u8 + num) as char
    } else {
        ('a' as u8 + num - 10) as char
    };
    s.push(ch)
}

fn append_hex2<const N : usize>(s : &mut String<N>, num : u8) -> Result<(), ()> {
    append_hex1(s, num >> 4)?;
    append_hex1(s, num & 0xf)
}

#[derive(PartialEq, Eq, Copy, Clone, Debug)]
#[cfg_attr(feature = "defmt", derive(defmt::Format))]
pub enum ScanResult {
    NAck,
    Ack,
    Error
}

// adapted from arduino-pico:
// https://github.com/earlephilhower/arduino-pico/blob/6e52b72523b11470b2ad2b0578ca1500be238108/libraries/Wire/src/Wire.cpp#L257
// This seems to be the only way to do zero-length writes:
// see https://github.com/raspberrypi/pico-sdk/issues/238
async fn handle_clock_stretching<'d, SCL: gpio::Pin>(scl: &mut Flex<'d, SCL>) -> Result<(), i2c::Error> {
    use i2c::Error::*;
    use i2c::AbortReason::*;

    if scl.is_high() {
        return Ok(())
    }

    let is_high = scl.wait_for_high();
    let timeout = Timer::after(Duration::from_micros(100));

    return match select(is_high, timeout).await {
        Either::First(_) => Ok(()),
        Either::Second(_) => Err(Abort(ArbitrationLoss)),
    }
}
async fn i2c_write_zero_bytes_middle_part<'d, SCL: gpio::Pin, SDA: gpio::Pin, const FREQ : u64>(
        mut addr: u16, delay: Duration, scl: &mut Flex<'d, SCL>, sda: &mut Flex<'d, SDA>)
        -> Result<(), i2c::Error> {
    sda.set_as_input();
    Timer::after(delay).await;
    scl.set_as_input();
    handle_clock_stretching(scl).await?;
    sda.set_as_output();
    Timer::after(delay).await;
    scl.set_as_output();
    Timer::after(delay).await;
    for _ in 0..8 {
        addr <<= 1;
        sda.set_level(if (addr & (1 << 7)) != 0 { gpio::Level::High } else { gpio::Level::Low });
        Timer::after(delay).await;
        scl.set_as_input();
        Timer::after(delay).await;
        handle_clock_stretching(scl).await?;
        scl.set_as_output();
        Timer::after(Duration::from_micros(5)).await; // Ensure we don't change too close to clock edge
    }

    sda.set_as_input();
    Timer::after(delay).await;
    scl.set_as_input();
    handle_clock_stretching(scl).await?;

    let result = if sda.is_low() {
        Ok(())
    } else {
        use i2c::Error::*;
        use i2c::AbortReason::*;
        Err(Abort(NoAcknowledge))
    };
    Timer::after(delay).await;
    scl.set_as_output();

    result
}
async fn i2c_write_zero_bytes<SCL: gpio::Pin, SDA: gpio::Pin, const FREQ : u64>(addr: u16, scl: &mut SCL, sda: &mut SDA)
        -> Result<(), i2c::Error> {
    use i2c::Error::*;
    use i2c::AbortReason::*;

    let delay = Duration::from_micros((1000000 / FREQ) / 2);

    let mut sda = Flex::new(sda);
    let mut scl = Flex::new(scl);
    sda.set_as_input();
    sda.set_pull(Pull::Up);
    scl.set_as_input();
    scl.set_pull(Pull::Up);
    // We only need low or pullup so we can set the value to low and only toggle direction.
    // This is how OutputOpenDrain does it (and the only reason we use Flex is because we want the pullup).
    sda.set_low();
    scl.set_low();

    Timer::after(delay).await;

    // abort if SCL is stuck low
    handle_clock_stretching(&mut scl).await?;

    // handle SDA being stuck low
    // (This can happen for very simple I2C devices that only use SCL and don't have any other clocks that
    // they can use to handle timeouts, e.g. simple port expanders.)
    if sda.is_low() {
        // I think I have read that DW_apb_i2c does nine clocks. This seems resonable so let's do the same.
        for _ in 0..9 {
            scl.set_as_output();
            Timer::after(delay).await;
            scl.set_as_input();
            Timer::after(delay).await;

            if sda.is_high() {
                break;
            }
        }

        if sda.is_low() || scl.is_low() {
            // There is isn't much that we can do here. A start condition would reset the state but we cannot
            // generate it while one of the lines is stuck low.
            return Err(Abort(ArbitrationLoss))
        }

        // SDA is high, again. Good. We will generate a start condition soon. This should abort any
        // partial transfers that some devices might think are ongoing. We have already waited for half
        // a clock so we are good to go.
    }

    let result = i2c_write_zero_bytes_middle_part::<_, _, FREQ>(addr, delay, &mut scl, &mut sda).await;

    Timer::after(delay).await;
    sda.set_as_output();
    Timer::after(delay).await;
    scl.set_as_input();
    Timer::after(delay).await;
    sda.set_as_input();
    Timer::after(delay).await;

    return result;
}

#[derive(PartialEq, Eq, Copy, Clone, Debug)]
#[cfg_attr(feature = "defmt", derive(defmt::Format))]
pub struct ScanResultForBus([ScanResult; 128]);

pub const MAX_ADDRESS : u8 = 128;
pub const SCAN_RESULT_STRING_LENGTH : usize = (MAX_ADDRESS as usize)*3 + (MAX_ADDRESS as usize)/16*6;

impl ScanResultForBus {
    pub fn new() -> ScanResultForBus {
        ScanResultForBus([ScanResult::Error; MAX_ADDRESS as usize])
    }

    pub async fn scan<SCL: gpio::Pin, SDA: gpio::Pin>(scl: &mut SCL, sda: &mut SDA) -> ScanResultForBus {
        let mut results = Self::new();
        results.scan_i2c(scl, sda).await;
        return results;
    }

    pub async fn scan_i2c<SCL: gpio::Pin, SDA: gpio::Pin>(self: &mut Self, scl: &mut SCL, sda: &mut SDA) {
        use i2c::Error::*;
        use i2c::AbortReason::*;

        for i in 0..MAX_ADDRESS {
            self.0[i as usize] = match i2c_write_zero_bytes::<_, _, 400_000>(i as u16, scl, sda).await {
                Result::Ok(()) => ScanResult::Ack,
                Result::Err(Abort(NoAcknowledge)) => ScanResult::NAck,
                Result::Err(e) => {
                    warn!("I2C problem: {:02x}: {:?}", i, e);
                    ScanResult::Error
                }
            }
        }
    }    

    pub fn format<const N : usize>(self: &Self, msg : &mut String<N>) -> Result<(), ()> {
        self::assert!(N >= SCAN_RESULT_STRING_LENGTH);

        for i in 0..MAX_ADDRESS/16 {
            let base = i*16;
            append_hex2(msg, base)?;
            msg.push_str(":")?;
            for j in 0..16 {
                msg.push_str(" ")?;
                match self.0[(base+j) as usize] {
                    ScanResult::Ack => append_hex2(msg, base+j)?,
                    ScanResult::NAck => msg.push_str("--")?,
                    ScanResult::Error => msg.push_str("!!")?,
                }
            }
            msg.push_str("\n")?;
        }
        Ok(())
    }
}

/*
impl PartialEq for ScanResultForBus {
    fn eq(self: &Self, other: &Self) -> bool {
        self.0.iter().zip(other.0.iter()).all(|(x,y)| x == y)
    }
}

impl Eq for ScanResultForBus { }
 */
